
//ENONCE:
/*Supposons disposer d’un nombre fini, MAX_RES, d’exemplaires de ressources d’un seul type.
Les processus peuvent demander un certain nombre, count, d’exemplaires parmi les
ressources disponibles (count <=MAX RES) pour les utiliser en exclusivité ; et puis une fois
terminé, ils doivent les retourner. Si, lors d’une demande d’allocation, le nombre d'exemplaires
demandés n’est pas disponible, le processus doit attendre qu'un nombre suffisant d'exemplaires
aient été libérés.

==> Proposez une solution de synchronisation par variables conditionnelles
*/

#include <semaphore.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>


#define MAX_RES 10
//the number of available resources
int available_res = MAX_RES;

pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t cond = PTHREAD_COND_INITIALIZER;

//Allocate_Res decreases available resources by count resources

int Allocate_Res(int count) 
{
	pthread_mutex_lock (&mutex);
	while (available_res < count)
	{
		// blocage et libération du mutex, compétition pour avoir le mutex s'il est réveillé
		pthread_cond_wait(&cond, &mutex); 
	}	
	// Here the mutex is locked even if the thread was executing the code inside "while"
	available_res -= count;
	pthread_mutex_unlock (&mutex);
}

/* Free_Res increases available resources by count */
int Free_Res(int count) 
{
	pthread_mutex_lock (&mutex);
	available_res += count;
	pthread_mutex_unlock (&mutex);
	pthread_cond_broadcast(&cond);
}

void * AllocFree()
{
	int count,i=0;

	while (i<5)
	{
		count=rand()%10;
		
		printf(" thread %lu needs %d Resources\n", pthread_self(),count);	
		Allocate_Res(count);
		printf(" thread %lu gets %d Resources---\n", pthread_self(),count);
		sleep(3);// use Res
		Free_Res(count);
		printf(" thread %lu releases %d Resources+++.\n", pthread_self(),count);
		i++;	
	}
}

int main()
{
	pthread_t thread1, thread2, thread3;
	pthread_create (&thread1, NULL, AllocFree, NULL);
	pthread_create (&thread2, NULL, AllocFree, NULL);
	pthread_create (&thread3, NULL, AllocFree, NULL);
	pthread_join(thread1, NULL);
	pthread_join(thread2, NULL);
	pthread_join(thread3, NULL);
	
}
